@@ -1,12 +1,15 @@
 /*
  * w - show what logged in users are doing.
  *
- * Almost entirely rewritten from scratch by Charles Blake circa
- * June 1996. Some vestigal traces of the original may exist.
- * That was done in 1993 by Larry Greenfield with some fixes by
- * Michael K. Johnson.
+ * Copyright © 2009-2023 Craig Small <csmall@dropbear.xyz>
+ * Copyright © 2011-2023 Jim Warner <james.warner@comcast.net>
+ * Copyright © 2011-2012 Sami Kerola <kerolasa@iki.fi>
+ * Copyright © 2002-2006 Albert Cahalan
+ * Copyright © 1996      Charles Blake
  *
- * Changes by Albert Cahalan, 2002.
+ * Rewritten, older version:
+ * Copyright © 1993      Larry Greenfield
+ *                       with some fixes by Michael K. Johnson.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -29,8 +32,6 @@
 #include <getopt.h>
 #include <limits.h>
 #include <locale.h>
-#include <locale.h>
-#include <pwd.h>
 #include <pwd.h>
 #include <signal.h>
 #include <stdio.h>
@@ -44,11 +45,22 @@
 #include <time.h>
 #include <unistd.h>
 #ifdef HAVE_UTMPX_H
-#	include <utmpx.h>
+#include <utmpx.h>
+#ifndef HAVE_UT_HOSTSIZE_IN_UTMPX
+#include <utmp.h>
+#endif
 #else
 #	include <utmp.h>
 #endif
 #include <arpa/inet.h>
+#ifdef WITH_SYSTEMD
+#      include <systemd/sd-login.h>
+#      include <systemd/sd-daemon.h>
+#endif
+#ifdef WITH_ELOGIND
+#      include <elogind/sd-login.h>
+#      include <elogind/sd-daemon.h>
+#endif
 
 #include "c.h"
 #include "fileutils.h"
@@ -66,11 +78,13 @@ typedef struct utmpx utmp_t;
 typedef struct utmp utmp_t;
 #endif
 
+#ifdef __GLIBC__
 #if !defined(UT_HOSTSIZE) || defined(__UT_HOSTSIZE)
 #	define UT_HOSTSIZE __UT_HOSTSIZE
 #	define UT_LINESIZE __UT_LINESIZE
 #	define UT_NAMESIZE __UT_NAMESIZE
 #endif
+#endif
 
 #ifdef W_SHOWFROM
 # define FROM_STRING "on"
@@ -192,7 +206,25 @@ static void print_display_or_interface(const char *restrict host, int len, int r
 
 
 /* This routine prints either the hostname or the IP address of the remote */
-static void print_from(const utmp_t *restrict const u, const int ip_addresses, const int fromlen) {
+static void print_from(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+		       const char *session,
+#endif
+		       const utmp_t *restrict const u, const int ip_addresses, const int fromlen) {
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+        if (session) {
+	    char *host = NULL;
+	    int r;
+
+	    r = sd_session_get_remote_host(session, &host);
+	    if (r < 0 || host == NULL)
+	        print_host("", 0, fromlen);
+	    else {
+	        print_host(host, strlen(host), fromlen);
+		free(host);
+	    }
+	} else {
+#endif
 	char buf[fromlen + 1];
 	char buf_ipv6[INET6_ADDRSTRLEN];
 	int len;
@@ -237,6 +269,9 @@ static void print_from(const utmp_t *restrict const u, const int ip_addresses, c
 #else
 	print_host(u->ut_host, UT_HOSTSIZE, fromlen);
 #endif
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+	}
+#endif
 }
 
 
@@ -251,34 +286,34 @@ static void print_time_ival7(time_t t, int centi_sec, FILE * fout)
 	if (oldstyle) {
 		if (t >= 48 * 60 * 60)
 			/* > 2 days */
-			fprintf(fout, _(" %2ludays"), t / (24 * 60 * 60));
+			fprintf(fout, _(" %2lludays"), (unsigned long long)t / (24 * 60 * 60));
 		else if (t >= 60 * 60)
 			/* > 1 hour */
 		        /* Translation Hint: Hours:Minutes */
-			fprintf(fout, " %2lu:%02u ", t / (60 * 60),
+			fprintf(fout, " %2llu:%02u ", (unsigned long long)t / (60 * 60),
 				(unsigned)((t / 60) % 60));
 		else if (t > 60)
 			/* > 1 minute */
 		        /* Translation Hint: Minutes:Seconds */
-			fprintf(fout, _(" %2lu:%02um"), t / 60, (unsigned)t % 60);
+			fprintf(fout, _(" %2llu:%02um"), (unsigned long long)t / 60, (unsigned)t % 60);
 		else
 			fprintf(fout, "       ");
 	} else {
 		if (t >= 48 * 60 * 60)
 			/* 2 days or more */
-			fprintf(fout, _(" %2ludays"), t / (24 * 60 * 60));
+			fprintf(fout, _(" %2lludays"), (unsigned long long)t / (24 * 60 * 60));
 		else if (t >= 60 * 60)
 			/* 1 hour or more */
 		        /* Translation Hint: Hours:Minutes */
-			fprintf(fout, _(" %2lu:%02um"), t / (60 * 60),
+			fprintf(fout, _(" %2llu:%02um"), (unsigned long long)t / (60 * 60),
 				(unsigned)((t / 60) % 60));
 		else if (t > 60)
 			/* 1 minute or more */
 		        /* Translation Hint: Minutes:Seconds */
-			fprintf(fout, " %2lu:%02u ", t / 60, (unsigned)t % 60);
+			fprintf(fout, " %2llu:%02u ", (unsigned long long)t / 60, (unsigned)t % 60);
 		else
 		        /* Translation Hint: Seconds:Centiseconds */
-			fprintf(fout, _(" %2lu.%02us"), t, centi_sec);
+			fprintf(fout, _(" %2llu.%02us"), (unsigned long long)t, centi_sec);
 	}
 }
 
@@ -339,7 +374,7 @@ static int get_tty_device(const char *restrict const name)
 
     for (i=0; dev_paths[i] != NULL; i++) {
         snprintf(buf, 32, dev_paths[i], name);
-        if (stat(buf, &st) == 0)
+        if (stat(buf, &st) == 0 && (st.st_mode & S_IFMT) == S_IFCHR)
             return st.st_rdev;
     }
     return -1;
@@ -353,17 +388,22 @@ static int get_tty_device(const char *restrict const name)
  * essential core of 'w'.
  */
 static int find_best_proc(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+        const char *session,
+#endif
         const utmp_t * restrict const u,
         const char *restrict const tty,
         unsigned long long *restrict const jcpu,
         unsigned long long *restrict const pcpu,
-        char *cmdline)
+        char *cmdline,
+        pid_t *pid)
 {
 #define PIDS_GETINT(e) PIDS_VAL(EU_ ## e, s_int, reap->stacks[i], info)
 #define PIDS_GETUNT(e) PIDS_VAL(EU_ ## e, u_int, reap->stacks[i], info)
 #define PIDS_GETULL(e) PIDS_VAL(EU_ ## e, ull_int, reap->stacks[i], info)
 #define PIDS_GETSTR(e) PIDS_VAL(EU_ ## e, str, reap->stacks[i], info)
     unsigned uid = ~0U;
+    pid_t ut_pid = -1;
     int found_utpid = 0;
     int i, total_procs, line;
     unsigned long long best_time = 0;
@@ -372,6 +412,7 @@ static int find_best_proc(
     struct pids_info *info=NULL;
     struct pids_fetch *reap;
     enum pids_item items[] = {
+        PIDS_ID_PID,
         PIDS_ID_TGID,
         PIDS_TICS_BEGAN,
         PIDS_ID_EUID,
@@ -382,12 +423,18 @@ static int find_best_proc(
         PIDS_TICS_ALL,
         PIDS_CMDLINE};
     enum rel_items {
-        EU_TGID, EU_START, EU_EUID, EU_RUID, EU_TPGID, EU_PGRP, EU_TTY,
+        EU_PID, EU_TGID, EU_START, EU_EUID, EU_RUID, EU_TPGID, EU_PGRP, EU_TTY,
         EU_TICS_ALL, EU_CMDLINE};
 
     *jcpu = 0;
     *pcpu = 0;
     if (!ignoreuser) {
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+        if (session) {
+            if (sd_session_get_uid(session, &uid) < 0)
+                return 0;
+        } else {
+#endif
         char buf[UT_NAMESIZE + 1];
         struct passwd *passwd_data;
         strncpy(buf, u->ut_user, UT_NAMESIZE);
@@ -396,11 +443,14 @@ static int find_best_proc(
             return 0;
         uid = passwd_data->pw_uid;
         /* OK to have passwd_data go out of scope here */
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+	}
+#endif
     }
 
     line = get_tty_device(tty);
 
-    if (procps_pids_new(&info, items, 9) < 0)
+    if (procps_pids_new(&info, items, 10) < 0)
         xerrx(EXIT_FAILURE,
               _("Unable to create pid info structure"));
     if ((reap = procps_pids_reap(info, PIDS_FETCH_TASKS_ONLY)) == NULL)
@@ -408,13 +458,21 @@ static int find_best_proc(
               _("Unable to load process information"));
     total_procs = reap->counts->total;
 
+    if (u)
+        ut_pid = u->ut_pid;
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+    else
+        sd_session_get_leader(session, &ut_pid);
+#endif
+
     for (i=0; i < total_procs; i++) {
         /* is this the login process? */
-        if (PIDS_GETINT(TGID) == u->ut_pid) {
+        if (PIDS_GETINT(TGID) == ut_pid) {
             found_utpid = 1;
             if (!best_time) {
                 best_time = PIDS_GETULL(START);
                 strncpy(cmdline, PIDS_GETSTR(CMDLINE), MAX_CMD_WIDTH);
+                *pid = PIDS_GETULL(PID);
                 *pcpu = PIDS_GETULL(TICS_ALL);
             }
 
@@ -426,6 +484,7 @@ static int find_best_proc(
             secondbest_time = PIDS_GETULL(START);
             if (cmdline[0] == '-' && cmdline[1] == '\0') {
                 strncpy(cmdline, PIDS_GETSTR(CMDLINE), MAX_CMD_WIDTH);
+                *pid = PIDS_GETULL(PID);
                 *pcpu = PIDS_GETULL(TICS_ALL);
             }
         }
@@ -438,6 +497,7 @@ static int find_best_proc(
             continue;
         best_time = PIDS_GETULL(START);
         strncpy(cmdline, PIDS_GETSTR(CMDLINE), MAX_CMD_WIDTH);
+        *pid = PIDS_GETULL(PID);
         *pcpu = PIDS_GETULL(TICS_ALL);
     }
     procps_pids_unref(&info);
@@ -449,26 +509,55 @@ static int find_best_proc(
 }
 
 static void showinfo(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+            const char *session, const char *name,
+#endif
             utmp_t * u, int formtype, int maxcmd, int from,
-            const int userlen, const int fromlen, const int ip_addresses)
+            const int userlen, const int fromlen, const int ip_addresses,
+            const int pids)
 {
     unsigned long long jcpu, pcpu;
     unsigned i;
     char uname[UT_NAMESIZE + 1] = "", tty[5 + UT_LINESIZE + 1] = "/dev/";
     long hertz;
     char cmdline[MAX_CMD_WIDTH + 1];
+    pid_t best_pid = -1;
+    int pids_length = 0;
 
     strcpy(cmdline, "-");
 
     hertz = procps_hertz_get();
+
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+    if (session) {
+        char *sd_tty;
+
+        if (sd_session_get_tty(session, &sd_tty) >= 0) {
+            for (i = 0; i < strlen (sd_tty); i++)
+                /* clean up tty if garbled */
+	        if (isalnum(sd_tty[i]) || (sd_tty[i] == '/'))
+		    tty[i + 5] = sd_tty[i];
+		else
+		    tty[i + 5] = '\0';
+	    free(sd_tty);
+	}
+    } else {
+#endif
     for (i = 0; i < UT_LINESIZE; i++)
         /* clean up tty if garbled */
         if (isalnum(u->ut_line[i]) || (u->ut_line[i] == '/'))
             tty[i + 5] = u->ut_line[i];
         else
             tty[i + 5] = '\0';
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+    }
+#endif
 
-    if (find_best_proc(u, tty + 5, &jcpu, &pcpu, cmdline) == 0)
+    if (find_best_proc(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+		       session,
+#endif
+		       u, tty + 5, &jcpu, &pcpu, cmdline, &best_pid) == 0)
     /*
      * just skip if stale utmp entry (i.e. login proc doesn't
      * exist). If there is a desire a cmdline flag could be
@@ -477,19 +566,44 @@ static void showinfo(
      */
         return;
 
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+    if (name)
+      strncpy(uname, name, UT_NAMESIZE);
+    else
+#endif
+      strncpy(uname, u->ut_user, UT_NAMESIZE);
     /* force NUL term for printf */
-    strncpy(uname, u->ut_user, UT_NAMESIZE);
+    uname[UT_NAMESIZE] = '\0';
 
     if (formtype) {
-        printf("%-*.*s%-9.8s", userlen + 1, userlen, uname, u->ut_line);
-        if (from)
-            print_from(u, ip_addresses, fromlen);
-#ifdef HAVE_UTMPX_H
-        print_logintime(u->ut_tv.tv_sec, stdout);
-#else
-        print_logintime(u->ut_time, stdout);
+        printf("%-*.*s%-9.8s", userlen + 1, userlen, uname, tty + 5);
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+        if (session) {
+            uint64_t ltime;
+
+            if (from)
+              print_from(session, NULL, ip_addresses, fromlen);
+
+            sd_session_get_start_time(session, &ltime);
+            print_logintime(ltime/((uint64_t) 1000000ULL), stdout);
+        } else {
 #endif
-        if (*u->ut_line == ':')
+            if (from)
+                print_from(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+			   NULL,
+#endif
+			   u, ip_addresses, fromlen);
+
+#ifdef HAVE_UTMPX_H
+            print_logintime(u->ut_tv.tv_sec, stdout);
+#else
+            print_logintime(u->ut_time, stdout);
+#endif
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+        }
+#endif
+        if (u && *u->ut_line == ':')
             /* idle unknown for xdm logins */
             printf(" ?xdm? ");
         else
@@ -503,16 +617,34 @@ static void showinfo(
         else
             printf("   ?   ");
     } else {
-        printf("%-*.*s%-9.8s", userlen + 1, userlen, u->ut_user,
-               u->ut_line);
+        printf("%-*.*s%-9.8s", userlen + 1, userlen, uname, tty + 5);
         if (from)
-            print_from(u, ip_addresses, fromlen);
-        if (*u->ut_line == ':')
+	    print_from(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+		       NULL,
+#endif
+		       u, ip_addresses, fromlen);
+        if (u && *u->ut_line == ':')
             /* idle unknown for xdm logins */
             printf(" ?xdm? ");
         else
             print_time_ival7(idletime(tty), 0, stdout);
     }
+    if (pids) {
+        pid_t ut_pid = -1;
+        if (u)
+	    ut_pid = u->ut_pid;
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+	else
+	    sd_session_get_leader(session, &ut_pid);
+#endif
+        pids_length = printf(" %d/%d", ut_pid, best_pid);
+        if (pids_length > maxcmd) {
+            maxcmd = 0;
+        } else if (pids_length > 0) {
+            maxcmd -= pids_length;
+        }
+    }
     printf(" %.*s\n", maxcmd, cmdline);
 }
 
@@ -529,6 +661,7 @@ static void __attribute__ ((__noreturn__))
 	fputs(_(" -f, --from          show remote hostname field\n"),out);
 	fputs(_(" -o, --old-style     old style output\n"),out);
 	fputs(_(" -i, --ip-addr       display IP address instead of hostname (if possible)\n"), out);
+	fputs(_(" -p, --pids          show the PID(s) of processes in WHAT\n"), out);
 	fputs(USAGE_SEPARATOR, out);
 	fputs(_("     --help     display this help and exit\n"), out);
 	fputs(USAGE_VERSION, out);
@@ -553,6 +686,7 @@ int main(int argc, char **argv)
 	int longform = 1;
 	int from = 1;
 	int ip_addresses = 0;
+	int pids = 0;
 
 	enum {
 		HELP_OPTION = CHAR_MAX + 1
@@ -565,6 +699,7 @@ int main(int argc, char **argv)
 		{"from", no_argument, NULL, 'f'},
 		{"old-style", no_argument, NULL, 'o'},
 		{"ip-addr", no_argument, NULL, 'i'},
+		{"pids", no_argument, NULL, 'p'},
 		{"help", no_argument, NULL, HELP_OPTION},
 		{"version", no_argument, NULL, 'V'},
 		{NULL, 0, NULL, 0}
@@ -583,7 +718,7 @@ int main(int argc, char **argv)
 #endif
 
 	while ((ch =
-		getopt_long(argc, argv, "husfoVi", longopts, NULL)) != -1)
+		getopt_long(argc, argv, "husfoVip", longopts, NULL)) != -1)
 		switch (ch) {
 		case 'h':
 			header = 0;
@@ -607,6 +742,9 @@ int main(int argc, char **argv)
 			ip_addresses = 1;
 			from = 1;
 			break;
+		case 'p':
+			pids = 1;
+			break;
 		case HELP_OPTION:
 			usage(stdout);
 		default:
@@ -666,7 +804,40 @@ int main(int argc, char **argv)
 		else
 			printf(_("   IDLE WHAT\n"));
 	}
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+	if (sd_booted() > 0) {
+		char **sessions_list;
+		int sessions;
+		int i;
 
+		sessions = sd_get_sessions (&sessions_list);
+		if (sessions < 0 && sessions != -ENOENT)
+			error(EXIT_FAILURE, -sessions, _("error getting sessions"));
+
+		if (sessions >= 0) {
+			for (int i = 0; i < sessions; i++) {
+				char *name;
+				int r;
+
+				if ((r = sd_session_get_username(sessions_list[i], &name)) < 0)
+					error(EXIT_FAILURE, -r, _("get user name failed"));
+
+				if (user) {
+					if (!strcmp(name, user))
+						showinfo(sessions_list[i], name, NULL, longform,
+							 maxcmd, from, userlen, fromlen,
+							 ip_addresses, pids);
+				} else {
+					showinfo(sessions_list[i], name, NULL, longform, maxcmd,
+						 from, userlen, fromlen, ip_addresses, pids);
+				}
+				free(name);
+				free(sessions_list[i]);
+			}
+			free(sessions_list);
+		}
+	} else {
+#endif
 #ifdef HAVE_UTMPX_H
 	setutxent();
 #else
@@ -685,8 +856,12 @@ int main(int argc, char **argv)
 			if (u->ut_type != USER_PROCESS)
 				continue;
 			if (!strncmp(u->ut_user, user, UT_NAMESIZE))
-				showinfo(u, longform, maxcmd, from, userlen,
-					 fromlen, ip_addresses);
+				showinfo(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+					 NULL, NULL,
+#endif
+					 u, longform, maxcmd, from, userlen,
+					 fromlen, ip_addresses, pids);
 		}
 	} else {
 		for (;;) {
@@ -700,8 +875,12 @@ int main(int argc, char **argv)
 			if (u->ut_type != USER_PROCESS)
 				continue;
 			if (*u->ut_user)
-				showinfo(u, longform, maxcmd, from, userlen,
-					 fromlen, ip_addresses);
+				showinfo(
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+					 NULL, NULL,
+#endif
+					 u, longform, maxcmd, from, userlen,
+					 fromlen, ip_addresses, pids);
 		}
 	}
 #ifdef HAVE_UTMPX_H
@@ -709,6 +888,9 @@ int main(int argc, char **argv)
 #else
 	endutent();
 #endif
+#if (defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)) && defined(HAVE_SD_SESSION_GET_LEADER)
+	}
+#endif
 
 	return EXIT_SUCCESS;
 }
